@@ -21,9 +21,15 @@
 
 #include <stdarg.h>
 #include <sys/param.h>
+#include <sys/ioctl.h>
+#include <fcntl.h>
 
 #include <linux/dm-ioctl.h>
 
+#ifdef linux
+#  include <linux/fs.h>
+#endif
+
 #ifdef HAVE_SELINUX
 #  include <selinux/selinux.h>
 #endif
@@ -118,6 +124,8 @@ struct dm_task *dm_task_create(int type)
 	dmt->gid = DEVICE_GID;
 	dmt->mode = DEVICE_MODE;
 	dmt->no_open_count = 0;
+	dmt->read_ahead = DM_READ_AHEAD_AUTO;
+	dmt->read_ahead_flags = 0;
 
 	return dmt;
 }
@@ -354,15 +362,119 @@ static int _rm_dev_node(const char *dev_name)
 	return 1;
 }
 
+#ifdef linux
+static int _open_dev_node(const char *dev_name)
+{
+	int fd = -1;
+	char path[PATH_MAX];
+
+	_build_dev_path(path, sizeof(path), dev_name);
+
+	if ((fd = open(path, O_RDONLY, 0)) < 0)
+		log_sys_error("open", path);
+
+	return fd;
+}
+
+int get_dev_node_read_ahead(const char *dev_name, uint32_t *read_ahead)
+{
+	int r = 1;
+	int fd;
+	long read_ahead_long;
+
+	if ((fd = _open_dev_node(dev_name)) < 0)
+		return_0;
+
+	if (ioctl(fd, BLKRAGET, &read_ahead_long)) {
+		log_sys_error("BLKRAGET", dev_name);
+		*read_ahead = 0;
+		r = 0;
+	}  else {
+		*read_ahead = (uint32_t) read_ahead_long;
+		log_debug("%s: read ahead is %" PRIu32, dev_name, *read_ahead);
+	}
+
+	if (close(fd))
+		stack;
+
+	return r;
+}
+
+static int _set_read_ahead(const char *dev_name, uint32_t read_ahead)
+{
+	int r = 1;
+	int fd;
+	long read_ahead_long = (long) read_ahead;
+
+	if ((fd = _open_dev_node(dev_name)) < 0)
+		return_0;
+
+	log_debug("%s: Setting read ahead to %" PRIu32, dev_name, read_ahead);
+
+	if (ioctl(fd, BLKRASET, read_ahead_long)) {
+		log_sys_error("BLKRASET", dev_name);
+		r = 0;
+	}
+
+	if (close(fd))
+		stack;
+
+	return r;
+}
+
+static int _set_dev_node_read_ahead(const char *dev_name, uint32_t read_ahead,
+				    uint32_t read_ahead_flags)
+{
+	uint32_t current_read_ahead;
+
+	if (read_ahead == DM_READ_AHEAD_AUTO)
+		return 1;
+
+	if (read_ahead == DM_READ_AHEAD_NONE)
+		read_ahead = 0;
+
+	if (read_ahead_flags & DM_READ_AHEAD_MINIMUM_FLAG) {
+		if (!get_dev_node_read_ahead(dev_name, &current_read_ahead))
+			return_0;
+
+		if (current_read_ahead > read_ahead) {
+			log_debug("%s: read ahead %" PRIu32
+				  " below minimum of %" PRIu32,
+				  dev_name, current_read_ahead, read_ahead);
+			return 1;
+		}
+	}
+
+	return _set_read_ahead(dev_name, read_ahead);
+}
+
+#else
+
+int get_dev_node_read_ahead(const char *dev_name, uint32_t *read_ahead)
+{
+	*read_ahead = 0;
+
+	return 1;
+}
+
+static int _set_dev_node_read_ahead(const char *dev_name, uint32_t read_ahead,
+				    uint32_t read_ahead_flags)
+{
+	return 1;
+}
+#endif
+
 typedef enum {
 	NODE_ADD,
 	NODE_DEL,
-	NODE_RENAME
+	NODE_RENAME,
+	NODE_READ_AHEAD
 } node_op_t;
 
 static int _do_node_op(node_op_t type, const char *dev_name, uint32_t major,
 		       uint32_t minor, uid_t uid, gid_t gid, mode_t mode,
-		       const char *old_name)
+		       const char *old_name, uint32_t read_ahead,
+		       uint32_t read_ahead_flags)
 {
 	switch (type) {
 	case NODE_ADD:
@@ -371,6 +483,9 @@ static int _do_node_op(node_op_t type, const char *dev_name, uint32_t major,
 		return _rm_dev_node(dev_name);
 	case NODE_RENAME:
 		return _rename_dev_node(old_name, dev_name);
+	case NODE_READ_AHEAD:
+		return _set_dev_node_read_ahead(dev_name, read_ahead,
+						read_ahead_flags);
 	}
 
 	return 1;
@@ -387,6 +502,8 @@ struct node_op_parms {
 	uid_t uid;
 	gid_t gid;
 	mode_t mode;
+	uint32_t read_ahead;
+	uint32_t read_ahead_flags;
 	char *old_name;
 	char names[0];
 };
@@ -400,7 +517,8 @@ static void _store_str(char **pos, char **ptr, const char *str)
 
 static int _stack_node_op(node_op_t type, const char *dev_name, uint32_t major,
 			  uint32_t minor, uid_t uid, gid_t gid, mode_t mode,
-			  const char *old_name)
+			  const char *old_name, uint32_t read_ahead,
+			  uint32_t read_ahead_flags)
 {
 	struct node_op_parms *nop;
 	size_t len = strlen(dev_name) + strlen(old_name) + 2;
@@ -418,6 +536,8 @@ static int _stack_node_op(node_op_t type, const char *dev_name, uint32_t major,
 	nop->uid = uid;
 	nop->gid = gid;
 	nop->mode = mode;
+	nop->read_ahead = read_ahead;
+	nop->read_ahead_flags = read_ahead_flags;
 
 	_store_str(&pos, &nop->dev_name, dev_name);
 	_store_str(&pos, &nop->old_name, old_name);
@@ -435,7 +555,8 @@ static void _pop_node_ops(void)
 	list_iterate_safe(noph, nopht, &_node_ops) {
 		nop = list_item(noph, struct node_op_parms);
 		_do_node_op(nop->type, nop->dev_name, nop->major, nop->minor,
-			    nop->uid, nop->gid, nop->mode, nop->old_name);
+			    nop->uid, nop->gid, nop->mode, nop->old_name,
+			    nop->read_ahead, nop->read_ahead_flags);
 		list_del(&nop->list);
 		dm_free(nop);
 	}
@@ -445,17 +566,28 @@ int add_dev_node(const char *dev_name, uint32_t major, uint32_t minor,
 		 uid_t uid, gid_t gid, mode_t mode)
 {
 	return _stack_node_op(NODE_ADD, dev_name, major, minor, uid, gid, mode,
-			      "");
+			      "", 0, 0);
 }
 
 int rename_dev_node(const char *old_name, const char *new_name)
 {
-	return _stack_node_op(NODE_RENAME, new_name, 0, 0, 0, 0, 0, old_name);
+	return _stack_node_op(NODE_RENAME, new_name, 0, 0, 0, 0, 0, old_name,
+			      0, 0);
 }
 
 int rm_dev_node(const char *dev_name)
 {
-	return _stack_node_op(NODE_DEL, dev_name, 0, 0, 0, 0, 0, "");
+	return _stack_node_op(NODE_DEL, dev_name, 0, 0, 0, 0, 0, "", 0, 0);
+}
+
+int set_dev_node_read_ahead(const char *dev_name, uint32_t read_ahead,
+			    uint32_t read_ahead_flags)
+{
+	if (read_ahead == DM_READ_AHEAD_AUTO)
+		return 1;
+
+	return _stack_node_op(NODE_READ_AHEAD, dev_name, 0, 0, 0, 0, 0, "",
+			      read_ahead, read_ahead_flags);
 }
 
 void update_devs(void)
@@ -463,9 +595,25 @@ void update_devs(void)
 	_pop_node_ops();
 }
 
-int dm_set_dev_dir(const char *dir)
+int dm_set_dev_dir(const char *dev_dir)
 {
-	snprintf(_dm_dir, sizeof(_dm_dir), "%s%s", dir, DM_DIR);
+	size_t len;
+	const char *slash;
+	if (*dev_dir != '/') {
+		log_debug("Invalid dev_dir value, %s: "
+			  "not an absolute name.", dev_dir);
+		return 0;
+	}
+
+	len = strlen(dev_dir);
+	slash = dev_dir[len-1] == '/' ? "" : "/";
+
+	if (snprintf(_dm_dir, sizeof _dm_dir, "%s%s%s", dev_dir, slash, DM_DIR)
+	    >= sizeof _dm_dir) {
+		log_debug("Invalid dev_dir value, %s: name too long.", dev_dir);
+		return 0;
+	}
+
 	return 1;
 }
 
